#define GC_REDIRECT_TO_LOCAL
#include <gc/gc.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <pthread.h>
#include "types.h"
#include "functions.h"
#include "int_funcs.h"

unsigned int gensym_last = 0;
pthread_mutex_t gensym_lock = PTHREAD_MUTEX_INITIALIZER;


#define ERROR_MSG_LENGTH	1024

#define SCM_ASSERT_ARG_CNT_EQ(__cnt,__fn_name) { \
	if (count != __cnt) {\
		char * errstr = (char *) GC_MALLOC(ERROR_MSG_LENGTH);	\
		sprintf(errstr, "Expecting %i arguments in '%s'\n", __cnt, __fn_name);\
		return scm_value_new_error(errstr); \
	}	\
}
#define __QUOTE_PREP(__x)	#__x

#define SCM_ASSERT_ARG_TYPE_EQ(__type,__position,__fn_name) { \
	if (SCM_TYPE(args[__position]) != __type) {\
		char * errstr = (char *) GC_MALLOC(ERROR_MSG_LENGTH);	\
		sprintf(errstr, "Expecting '%s' in arg. %i of '%s'\n", __QUOTE_PREP(__type), __position + 1, __fn_name);\
		return scm_value_new_error(errstr); \
	}	\
}

#define SCM_ASSERT_ARG_TYPE_NUMBER(__position,__fn_name) { \
	if ((SCM_TYPE(args[__position]) != INT) && (SCM_TYPE(args[__position]) != FLOAT)) {\
		char * errstr = (char *) GC_MALLOC(ERROR_MSG_LENGTH);	\
		sprintf(errstr, "Expecting 'number' in arg. %i of '%s'\n", __position + 1, __fn_name);\
		return scm_value_new_error(errstr); \
	}	\
}

#define scm_number(x) (SCM_TYPE(x) == INT ? SCM_INT(x) : SCM_FLOAT(x))

scm_value * plus(SCM_ARGS)
{
	int i, result_float = 0;
	for (i = 0; i < count; i++) {
		SCM_ASSERT_ARG_TYPE_NUMBER(i, "plus")
		if (SCM_TYPE(args[i]) == FLOAT) result_float = 1;
	}

	if (result_float) {
		double result = 0;
		for (i = 0; i < count; i++) 
			result += scm_number(args[i]);
		return scm_value_new_float(result);
	} else {
		int result = 0;
		for (i = 0; i < count; i++)
			result += scm_number(args[i]);
		return scm_value_new_int(result);
	}
}

scm_value * minus(SCM_ARGS)
{
	int i, result_float = 0;
	if (count == 0) return scm_value_new_int(0);
	if (count == 1) {
		SCM_ASSERT_ARG_TYPE_NUMBER(0, "minus")
		if (SCM_TYPE(args[0]) == INT) return scm_value_new_int(- SCM_INT(args[0]));
		if (SCM_TYPE(args[0]) == FLOAT) return scm_value_new_float(- SCM_FLOAT(args[0]));
	}

	for (i = 0; i < count; i++) {
		SCM_ASSERT_ARG_TYPE_NUMBER(i, "minus")
		if (SCM_TYPE(args[i]) == FLOAT) result_float = 1;
	}

	if (result_float) {
		double result = scm_number(args[0]);
		for (i = 1; i < count; i++) 
			result -= scm_number(args[i]);
		return scm_value_new_float(result);
	} else {
		int result = scm_number(args[0]);
		for (i = 1; i < count; i++) 
			result -= scm_number(args[i]);
		return scm_value_new_int(result);
	}
}

scm_value * multiply(SCM_ARGS)
{
	int i, result_float = 0;
	result_float = 0;
	for (i = 0; i < count; i++) {
		SCM_ASSERT_ARG_TYPE_NUMBER(i, "multiply")
		if (SCM_TYPE(args[i]) == FLOAT) result_float = 1;
	}

	for (i = 0; i < count; i++) {
		SCM_ASSERT_ARG_TYPE_NUMBER(i, "minus")
		if (SCM_TYPE(args[i]) == FLOAT) result_float = 1;
	}

	if (result_float) {
		double result = 1;
		for (i = 0; i < count; i++) 
			result *= scm_number(args[i]);
		return scm_value_new_float(result);
	} else {
		int result = 1;
		for (i = 0; i < count; i++) 
			result *= scm_number(args[i]);
		return scm_value_new_int(result);
	}
}

scm_value * divide(SCM_ARGS)
{
	int i, result_float;
	result_float = 0;
	for (i = 0; i < count; i++) {
		SCM_ASSERT_ARG_TYPE_NUMBER(i, "divide")
		if (SCM_TYPE(args[i]) == FLOAT) result_float = 1;
		if ((scm_number(args[i]) == 0) && (i > 0)) return scm_value_new_error("Division by zero"); 
	}
	if (count == 1) {
		if (scm_number(args[0]) == 0) return scm_value_new_error("Division by zero"); 
		if ((SCM_TYPE(args[0]) == INT) && (scm_number(args[0]) == 1)) return scm_value_new_int(1); 
		return scm_value_new_float(1 / scm_number(args[0]));
	}

	double result = scm_number(args[0]);
	for (i = 1; i < count; i++) 
		result /= scm_number(args[i]);
	return scm_value_new_float(result);
}

scm_value * quotient(SCM_ARGS)
{
	SCM_ASSERT_ARG_CNT_EQ(2, "quotient")
	SCM_ASSERT_ARG_TYPE_EQ(INT, 0, "quotient")
	SCM_ASSERT_ARG_TYPE_EQ(INT, 1, "quotient")
	
	return scm_value_new_int(SCM_INT(args[0]) / SCM_INT(args[1]));
}

scm_value * modulo(SCM_ARGS)
{
	SCM_ASSERT_ARG_CNT_EQ(2, "modulo")
	SCM_ASSERT_ARG_TYPE_EQ(INT, 0, "modulo")
	SCM_ASSERT_ARG_TYPE_EQ(INT, 1, "modulo")
	
	return scm_value_new_int(SCM_INT(args[0]) % SCM_INT(args[1]));
}

scm_value * pred_eqref(SCM_ARGS)
{
	SCM_ASSERT_ARG_CNT_EQ(2, "pred_eqref")
	if (args[0] == args[1]) return SCM_CONST_TRUE;
	return SCM_CONST_FALSE;
}

scm_value * pred_eq(SCM_ARGS)
{
	SCM_ASSERT_ARG_CNT_EQ(2, "pred_eq")
		/* FIXME: support for float */
	SCM_ASSERT_ARG_TYPE_EQ(INT, 0, "pred_eq")
	SCM_ASSERT_ARG_TYPE_EQ(INT, 1, "pred_eq")
	
	if (SCM_INT(args[0]) == SCM_INT(args[1])) return SCM_CONST_TRUE;
	return SCM_CONST_FALSE;
}


scm_value * pred_equal(SCM_ARGS)
{
	scm_value * a1;
	scm_value * a2;

	SCM_ASSERT_ARG_CNT_EQ(2, "pred_equal")
	
	a1 = args[0];
	a2 = args[1];
	if (SCM_TYPE(a1) != SCM_TYPE(a2)) return SCM_CONST_FALSE;
	switch (SCM_TYPE(a1)) {
		case INT: return (SCM_INT(a1) == SCM_INT(a2) ? SCM_CONST_TRUE : SCM_CONST_FALSE);
		case FLOAT: return (SCM_FLOAT(a1) == SCM_FLOAT(a2) ? SCM_CONST_TRUE : SCM_CONST_FALSE);
		case SYMBOL: return (!strcmp(SCM_SYMBOL(a1), SCM_SYMBOL(a2)) ? SCM_CONST_TRUE : SCM_CONST_FALSE);
		case BOOL: return (SCM_BOOL(a1) == SCM_BOOL(a2) ? SCM_CONST_TRUE : SCM_CONST_FALSE);
		case NIL: return SCM_CONST_TRUE;
		case PAIR:
			  {
				  scm_value * nargs[4];
				  scm_value * v1;
				  scm_value * v2;

				  nargs[0] = SCM_CAR(args[0]);
				  nargs[1] = SCM_CAR(args[1]);
				  nargs[2] = SCM_CDR(args[0]);
				  nargs[3] = SCM_CDR(args[1]);

				  v1 = pred_equal(2, nargs);
				  v2 = pred_equal(2, nargs + 2);

				  if (SCM_BOOL(v1) && SCM_BOOL(v2)) return SCM_CONST_TRUE;
				  return SCM_CONST_FALSE;
			  }
		case VECTOR: 
			  {
				  if (SCM_VECTOR(args[0])->size != SCM_VECTOR(args[1])->size) return SCM_CONST_FALSE;
				  scm_value * nargs[2];
				  int i;
				  for (i = 0; i < SCM_VECTOR(args[0])->size; i++) {
					  nargs[0] = SCM_VECTOR(args[0])->items[i];
					  nargs[1] = SCM_VECTOR(args[1])->items[i];
					  scm_value * v = pred_equal(2, nargs);
					  if (!SCM_BOOL(v)) return SCM_CONST_FALSE;
				  }
				  return SCM_CONST_TRUE;
			  }
	}
	return scm_value_new_bool(a1 == a2);
}

scm_value * pred_eqv(SCM_ARGS)
{
	scm_value * a1;
	scm_value * a2;

	SCM_ASSERT_ARG_CNT_EQ(2, "pred_equal")
	
	a1 = args[0];
	a2 = args[1];
	if (SCM_TYPE(a1) != SCM_TYPE(a2)) return SCM_CONST_FALSE;
	switch (SCM_TYPE(a1)) {
		case INT: return (SCM_INT(a1) == SCM_INT(a2) ? SCM_CONST_TRUE : SCM_CONST_FALSE);
		case FLOAT: return (SCM_FLOAT(a1) == SCM_FLOAT(a2) ? SCM_CONST_TRUE : SCM_CONST_FALSE);
		case SYMBOL: return (!strcmp(SCM_SYMBOL(a1), SCM_SYMBOL(a2)) ? SCM_CONST_TRUE : SCM_CONST_FALSE);
		case BOOL: return (SCM_BOOL(a1) == SCM_BOOL(a2) ? SCM_CONST_TRUE : SCM_CONST_FALSE);
		case NIL: return SCM_CONST_TRUE;
	}
	return scm_value_new_bool(a1 == a2);
}


scm_value * pred_lt(SCM_ARGS)
{
	SCM_ASSERT_ARG_CNT_EQ(2, "pred_lt")
	SCM_ASSERT_ARG_TYPE_NUMBER(0, "pred_lt")
	SCM_ASSERT_ARG_TYPE_NUMBER(1, "pred_lt")
	
	if (scm_number(args[0]) < scm_number(args[1])) return SCM_CONST_TRUE;
	return SCM_CONST_FALSE;
}

scm_value * pred_gt(SCM_ARGS)
{
	SCM_ASSERT_ARG_CNT_EQ(2, "pred_gt")
	SCM_ASSERT_ARG_TYPE_NUMBER(0, "pred_gt")
	SCM_ASSERT_ARG_TYPE_NUMBER(1, "pred_gt")
	
	if (scm_number(args[0]) > scm_number(args[1])) return SCM_CONST_TRUE;
	return SCM_CONST_FALSE;
}

#define TYPE_PREDICATE(fun_name,pred_type)	\
scm_value * fun_name(SCM_ARGS)	\
{	\
	SCM_ASSERT_ARG_CNT_EQ(1, "fun_name")	\
	if (SCM_TYPE(args[0]) == pred_type) return SCM_CONST_TRUE;	\
	else return SCM_CONST_FALSE;	\
}

TYPE_PREDICATE(pred_symbol, SYMBOL)
TYPE_PREDICATE(pred_boolean, BOOL)
TYPE_PREDICATE(pred_macro, MACRO)
TYPE_PREDICATE(pred_pair, PAIR)
TYPE_PREDICATE(pred_int, INT)
TYPE_PREDICATE(pred_float, FLOAT)
TYPE_PREDICATE(pred_vector, VECTOR)

scm_value * pred_procedure(SCM_ARGS)
{
	SCM_ASSERT_ARG_CNT_EQ(1, "pred_procedure")
	return scm_value_new_bool((SCM_TYPE(args[0]) == LAMBDA) || (SCM_TYPE(args[0]) == FUNC));
}

scm_value * pred_number(SCM_ARGS)
{
	SCM_ASSERT_ARG_CNT_EQ(1, "pred_procedure")
	return scm_value_new_bool((SCM_TYPE(args[0]) == INT) || (SCM_TYPE(args[0]) == FLOAT));
}

scm_value * pred_list(SCM_ARGS)
{
	scm_value * arg;
	SCM_ASSERT_ARG_CNT_EQ(1, "pred_list")
	arg = args[0];
	while (1) {
		if (SCM_TYPE(arg) == NIL) return SCM_CONST_TRUE;
		if (SCM_TYPE(arg) == PAIR) arg = SCM_CDR(arg);
		else return SCM_CONST_FALSE;
	}
}
/*
scm_value * display(SCM_ARGS)
{
	SCM_ASSERT_ARG_CNT_EQ(1, "display_value")
	display_value(args[0]);
	return SCM_CONST_VOID;
}
*/
scm_value * display_dbg(SCM_ARGS)
{
	SCM_ASSERT_ARG_CNT_EQ(1, "display_value_dbg")
	display_value_dbg(args[0]);
	return SCM_CONST_VOID;
}
/*
scm_value * newline(SCM_ARGS)
{
	SCM_ASSERT_ARG_CNT_EQ(0, "newline")
	printf("\n");
	return SCM_CONST_VOID;
}
*/
scm_value * list(SCM_ARGS)
{
	if (count == 0) return SCM_CONST_NIL;
	return scm_value_new_pair(args[0], list(count - 1, args + 1));
}

scm_value * is_null(SCM_ARGS)
{
	SCM_ASSERT_ARG_CNT_EQ(1, "is_null")

	if (SCM_TYPE(args[0]) == NIL) return SCM_CONST_TRUE;
	return SCM_CONST_FALSE;
}

scm_value * car(SCM_ARGS)
{
	SCM_ASSERT_ARG_CNT_EQ(1, "car")
	SCM_ASSERT_ARG_TYPE_EQ(PAIR, 0, "car");
	
	return SCM_CAR(args[0]);
}

scm_value * cdr(SCM_ARGS)
{
	SCM_ASSERT_ARG_CNT_EQ(1, "cdr")
	SCM_ASSERT_ARG_TYPE_EQ(PAIR, 0, "cdr");
	
	return SCM_CDR(args[0]);
}

scm_value * cons(SCM_ARGS)
{
	SCM_ASSERT_ARG_CNT_EQ(2, "cons")
	return scm_value_new_pair(args[0], args[1]);
}

scm_value * not(SCM_ARGS)
{
	SCM_ASSERT_ARG_CNT_EQ(1, "not")
	scm_value * result = SCM_CONST_FALSE;
	scm_value * val = args[0];
	if (SCM_TYPE(val) == BOOL) result = (SCM_BOOL(val) ? SCM_CONST_FALSE : SCM_CONST_TRUE);
	if (SCM_TYPE(val) == INT) result = (SCM_INT(val) ? SCM_CONST_FALSE : SCM_CONST_TRUE);
	if (SCM_TYPE(val) == FLOAT) result = (SCM_FLOAT(val) ? SCM_CONST_FALSE : SCM_CONST_TRUE);
	if (SCM_TYPE(val) == NIL) result = SCM_CONST_FALSE;

	return result;
}

scm_value * gensym(SCM_ARGS)
{
	unsigned int iid;
	char * id = (char *) GC_MALLOC(1024);
	
	SCM_ASSERT_ARG_CNT_EQ(0, "gensym");

	pthread_mutex_lock(&gensym_lock);
	iid = gensym_last++;
	pthread_mutex_unlock(&gensym_lock);

	//sprintf(id, "#;g %i#", iid);
	sprintf(id, "__gensym__%i", iid);
	return scm_value_new_symbol(id);
}

scm_value * append(SCM_ARGS)
{
	if (count == 0) return scm_value_new_nil();
	if (count == 1) return args[0];

	if (!is_list(args[0]) || !is_list(args[1])) return scm_value_new_error("'Append' arguments have to be lists");
	scm_value * result = args[1];
	scm_value * appended = list_reverse(args[0]);
	while (!SCM_IS_NULL(appended)) {
		result = scm_value_new_pair(SCM_CAR(appended), result);
		appended = SCM_CDR(appended);
	}
	
	return result;
}

scm_value * vector(SCM_ARGS)
{
	int i;
	scm_value * result = scm_value_new_vector(count, scm_value_new_nil());
	for (i = 0; i < count; i++)
		SCM_VECTOR(result)->items[i] = args[i];
	return result;
}

scm_value * make_vector(SCM_ARGS)
{
	scm_value * val;
	SCM_ASSERT_ARG_TYPE_EQ(INT, 0, "make-vector");
	if ((count != 1) && (count != 2)) return scm_value_new_error("'make-vector' requires 2 or 3 arguments");
	val = (count == 2 ? args[1] : scm_value_new_bool(0));
	return scm_value_new_vector(SCM_INT(args[0]), val);
}

scm_value * vector_ref(SCM_ARGS)
{
	SCM_ASSERT_ARG_CNT_EQ(2, "vector-ref");
	SCM_ASSERT_ARG_TYPE_EQ(VECTOR, 0, "vector-ref");
	SCM_ASSERT_ARG_TYPE_EQ(INT, 1, "vector-ref");

	scm_vector * vector = SCM_VECTOR(args[0]);
	if (SCM_INT(args[1]) >= vector->size) return scm_value_new_error("Value out of range in 'vector-ref'");
	if (SCM_INT(args[1]) < 0) return scm_value_new_error("Value out of range in 'vector-ref'");
	return vector->items[SCM_INT(args[1])];
}

scm_value * vector_length(SCM_ARGS)
{
	SCM_ASSERT_ARG_CNT_EQ(1, "vector-length");
	SCM_ASSERT_ARG_TYPE_EQ(VECTOR, 0, "vector-length");

	return scm_value_new_int(SCM_VECTOR(args[0])->size);
}

scm_value * list_to_vector(SCM_ARGS)
{
	SCM_ASSERT_ARG_CNT_EQ(1, "list->vector");
	if (!is_list(args[0])) return scm_value_new_error("'list->vector' argument have to be list");
	return scm_value_new_vector_from_list(args[0]);
}

scm_value * vector_to_list(SCM_ARGS)
{
	SCM_ASSERT_ARG_CNT_EQ(1, "vector->list");
	SCM_ASSERT_ARG_TYPE_EQ(VECTOR, 0, "vector->list");
	int i;
	scm_vector * vector = SCM_VECTOR(args[0]);
	scm_value * result = scm_value_new_nil();
	for (i = vector->size - 1; i >= 0; i--)
		result = scm_value_new_pair(vector->items[i], result);
	return result;
}

scm_value * begin(SCM_ARGS)
{
	int i;
	scm_value * result = scm_value_new_nil();
	for (i = count - 1; i >= 0; i--)
		result = scm_value_new_pair(args[i], result);

	result = scm_value_new_pair(
			scm_value_new_symbol("lambda"),
			scm_value_new_pair(scm_value_new_nil(), result));
	result = scm_value_new_pair(result, scm_value_new_nil());
	return result;
}

scm_value * let(SCM_ARGS)
{
	int i;
	if (count < 2) return scm_value_new_error("Syntax error in let - missing expression");
	scm_value * result = scm_value_new_nil();
	scm_value * variables = scm_value_new_nil();
	scm_value * values = scm_value_new_nil();
	scm_value * bindings;
	scm_value * named_let = NULL;

	if (SCM_TYPE(args[0]) == SYMBOL) {
		named_let = args[0];
		bindings = args[1];
	} else bindings = args[0];

	if (!is_list(bindings)) return scm_value_new_error("Syntax error in let - bad bindings");
	while (!SCM_IS_NULL(bindings)) {
		scm_value * assoc = SCM_CAR(bindings);
		if (!is_list(assoc)) return scm_value_new_error("Syntax error in let - bad binding");
		if (list_length(assoc) != 2) return scm_value_new_error("Syntax error in let - bad binding");

		variables = scm_value_new_pair(SCM_CAR(assoc), variables);
		values = scm_value_new_pair(SCM_CADR(assoc), values);
		bindings = SCM_CDR(bindings);
	}

	values = list_reverse(values);
	variables = list_reverse(variables);

	if (!named_let) {
		for (i = count - 1; i >= 1; i--) 
			result = scm_value_new_pair(args[i], result);

		result = scm_value_new_pair(
				scm_value_new_symbol("lambda"),
				scm_value_new_pair(variables, result));
		result = scm_value_new_pair(result, values);
	} else {
		scm_value * definition = scm_value_new_pair(named_let, variables);
		scm_value * call = scm_value_new_pair(named_let, values);
		for (i = count - 1; i >= 2; i--) 
			result = scm_value_new_pair(args[i], result);
		result = scm_value_new_pair(scm_value_new_symbol("define"), scm_value_new_pair(definition, result));
		result = scm_value_new_pair(scm_value_new_symbol("lambda"), scm_value_new_pair(scm_value_new_nil(), scm_value_new_pair(result, scm_value_new_pair(call, scm_value_new_nil()))));
		result = scm_value_new_pair(result, scm_value_new_nil());
	}
	return result;
}

#define __QUOTE_SYMBOL(arg) (scm_value_new_pair(scm_value_new_symbol("quote"), scm_value_new_pair(arg, scm_value_new_nil())))
#define __QUOTE_JOIN(op,arg1,arg2) (scm_value_new_pair(scm_value_new_symbol(op), \
			scm_value_new_pair(arg1, scm_value_new_pair(arg2, scm_value_new_nil()))))
scm_value * __quasiquote(scm_value * arg)
{
	scm_value * result = scm_value_new_error("Syntax error in quasiquote");
	int is_vector = 0;

	if (SCM_TYPE(arg) == VECTOR) {
		arg = vector_to_list(1, &arg);
		is_vector = 1;
	}
	if (SCM_TYPE(arg) == PAIR) {
		if (SCM_TYPE(SCM_CAR(arg)) == PAIR) {
			scm_value * sublist = SCM_CAAR(arg);
			if (SCM_TYPE(sublist) == SYMBOL) {
				if (!strcmp("unquote", SCM_SYMBOL(sublist)))
					result = __QUOTE_JOIN("cons", SCM_CADAR(arg), __quasiquote(SCM_CDR(arg)));
				if (!strcmp("unquote-splicing", SCM_SYMBOL(sublist)))
					result = __QUOTE_JOIN("append", SCM_CADAR(arg), __quasiquote(SCM_CDR(arg)));
			} else result = __QUOTE_JOIN("cons", __quasiquote(SCM_CAR(arg)), __quasiquote(SCM_CDR(arg)));
		} else result = __QUOTE_JOIN("cons", __quasiquote(SCM_CAR(arg)), __quasiquote(SCM_CDR(arg)));
	} else if (SCM_TYPE(arg) != VECTOR) result = __QUOTE_SYMBOL(arg);
	if (is_vector) {
		result = scm_value_new_pair(scm_value_new_symbol("list->vector"), scm_value_new_pair(result, scm_value_new_nil()));
	}
	return result;
}

scm_value * quasiquote(SCM_ARGS)
{
	SCM_ASSERT_ARG_CNT_EQ(1, "quasiquote");
	return __quasiquote(args[0]);
}

scm_value * and(SCM_ARGS)
{
	int i;
	scm_value * result = scm_value_new_symbol("#t");
	for (i = count - 1; i >= 0; i--) {
		result = scm_value_new_pair(scm_value_new_symbol("if"),
				scm_value_new_pair(args[i],
					scm_value_new_pair(result, 
					scm_value_new_pair(scm_value_new_symbol("#f"),
						scm_value_new_nil()))));
	}
	return result;
}

scm_value * or(SCM_ARGS)
{
	int i;
	scm_value * result = scm_value_new_symbol("#f");
	for (i = count - 1; i >= 0; i--) {
		result = scm_value_new_pair(scm_value_new_symbol("if"),
				scm_value_new_pair(args[i],
					scm_value_new_pair(scm_value_new_symbol("#t"), 
					scm_value_new_pair(result,
						scm_value_new_nil()))));
	}
	return result;
}
/*
scm_value * cond(SCM_ARGS)
{
	scm_value * result;
	scm_value * symbol;
	if (count == 0) return scm_value_new_void();
	scm_value * first = args[0];
	if (SCM_TYPE(first) != PAIR) return scm_value_new_error("Syntax error in cond");
	if ((SCM_TYPE(first) == SYMBOL) && (!strcmp(SCM_SYMBOL(SCM_CAR(first)), "else")))
		return SCM_CONS(scm_value_new_symbol("begin"), SCM_CDR(first));

	symbol = gensym(0, NULL);
	scm_value * assign = SCM_CONS(SCM_CONS(symbol, SCM_CONS(SCM_CAR(first), scm_value_new_nil())), scm_value_new_nil());

	scm_value * branch1 = scm_value_new_void();
	if (SCM_IS_NULL(SCM_CDR(first))) branch1 = symbol;

	if ((SCM_TYPE(SCM_CDR(first)) == PAIR) && (SCM_TYPE(SCM_CAR(SCM_CDR(first))) == SYMBOL)
		&& (!strcmp(SCM_SYMBOL(SCM_CAR(SCM_CDR(first))), "=>"))) {
		if (SCM_TYPE(SCM_CDR(SCM_CDR(first))) != PAIR) return scm_value_new_error("Syntax error in cond");
		branch1 = SCM_CONS(SCM_CAR(SCM_CDR(SCM_CDR(first))), SCM_CONS(symbol, scm_value_new_nil()));
	} else {
		branch1 = SCM_CONS(scm_value_new_symbol("begin"), SCM_CDR(first));
	}

	scm_value * branch2 = list(count - 1, args + 1);
	branch2 = SCM_CONS(scm_value_new_symbol("cond"), branch2);

	scm_value * body = SCM_CONS(scm_value_new_symbol("if"), SCM_CONS(symbol, SCM_CONS(branch1, SCM_CONS(branch2, scm_value_new_nil()))));

	result = SCM_CONS(scm_value_new_symbol("let"), SCM_CONS(assign, SCM_CONS(body, scm_value_new_nil())));


	return result;
}
*/
