#include "types.h"
#include "env.h"
#include "functions.h"

scm_env * get_initial_env(scm_env * parent) 
{
	scm_env * env = scm_env_new_with_hash(parent, 1000);
	
	/* special forms */
	scm_env_add(env, "define", scm_value_new_spec_form(SP_DEFINE));
	scm_env_add(env, "if", scm_value_new_spec_form(SP_IF));
	scm_env_add(env, "lambda", scm_value_new_spec_form(SP_LAMBDA));
	scm_env_add(env, "quote", scm_value_new_spec_form(SP_QUOTE));
	scm_env_add(env, "define-macro", scm_value_new_spec_form(SP_MACRO));
	scm_env_add(env, "set!", scm_value_new_spec_form(SP_SET));
	scm_env_add(env, "call/cc", scm_value_new_spec_form(SP_CALLCC));
	scm_env_add(env, "call-with-current-continuation", scm_value_new_spec_form(SP_CALLCC));
	scm_env_add(env, "map-exp", scm_value_new_spec_form(SP_MAP_EXP));
	scm_env_add(env, "set-car!", scm_value_new_spec_form(SP_SETCAR));
	scm_env_add(env, "set-cdr!", scm_value_new_spec_form(SP_SETCDR));
	scm_env_add(env, "vector-set!", scm_value_new_spec_form(SP_VECTOR_SET));

	/* precompiled macros */
	scm_env_add(env, "begin", scm_value_new_precompiled_macro(begin));
	scm_env_add(env, "let", scm_value_new_precompiled_macro(let));
	scm_env_add(env, "quasiquote", scm_value_new_precompiled_macro(quasiquote));
	scm_env_add(env, "and", scm_value_new_precompiled_macro(and));
	scm_env_add(env, "or", scm_value_new_precompiled_macro(or));
	
	/* numeric operations */
	scm_env_add(env, "+", scm_value_new_func(plus));
	scm_env_add(env, "-", scm_value_new_func(minus));
	scm_env_add(env, "*", scm_value_new_func(multiply));
	scm_env_add(env, "/", scm_value_new_func(divide));
	scm_env_add(env, "modulo", scm_value_new_func(modulo));
	scm_env_add(env, "quotient", scm_value_new_func(quotient));
	
	/* logic operations */
	scm_env_add(env, "=", scm_value_new_func(pred_eq));
	scm_env_add(env, "<", scm_value_new_func(pred_lt));
	scm_env_add(env, ">", scm_value_new_func(pred_gt));
	scm_env_add(env, "#f", SCM_CONST_FALSE);
	scm_env_add(env, "#t", SCM_CONST_TRUE);
	scm_env_add(env, "not", scm_value_new_func(not));

	scm_env_add(env, "pair?", scm_value_new_func(pred_pair));
	scm_env_add(env, "symbol?", scm_value_new_func(pred_symbol));
	scm_env_add(env, "boolean?", scm_value_new_func(pred_boolean));
	scm_env_add(env, "number?", scm_value_new_func(pred_number));
	scm_env_add(env, "procedure?", scm_value_new_func(pred_procedure));
	scm_env_add(env, "macro?", scm_value_new_func(pred_macro));
	scm_env_add(env, "list?", scm_value_new_func(pred_list));
	scm_env_add(env, "equal?", scm_value_new_func(pred_equal));
	scm_env_add(env, "eqv?", scm_value_new_func(pred_eqv));
	scm_env_add(env, "int?", scm_value_new_func(pred_int));
	scm_env_add(env, "float?", scm_value_new_func(pred_float));
	scm_env_add(env, "vector?", scm_value_new_func(pred_vector));
	scm_env_add(env, "eq?", scm_value_new_func(pred_eqref));

	/* vectors */
	scm_env_add(env, "vector", scm_value_new_func(vector));
	scm_env_add(env, "make-vector", scm_value_new_func(make_vector));
	scm_env_add(env, "vector-ref", scm_value_new_func(vector_ref));
	scm_env_add(env, "vector-length", scm_value_new_func(vector_length));
	scm_env_add(env, "vector->list", scm_value_new_func(vector_to_list));
	scm_env_add(env, "list->vector", scm_value_new_func(list_to_vector));

	/* lists */
	scm_env_add(env, "list", scm_value_new_func(list));
	scm_env_add(env, "null?", scm_value_new_func(is_null));
	scm_env_add(env, "car", scm_value_new_func(car));
	scm_env_add(env, "cdr", scm_value_new_func(cdr));
	scm_env_add(env, "cons", scm_value_new_func(cons));
	scm_env_add(env, "append", scm_value_new_func(append));

	/* output */
	scm_env_add(env, "newline", scm_value_new_spec_form(SP_NEWLINE));
	scm_env_add(env, "display", scm_value_new_spec_form(SP_DISPLAY));
	scm_env_add(env, "display-dbg", scm_value_new_func(display_dbg));

	scm_env_add(env, "gensym", scm_value_new_func(gensym));

	return env;
}
