#define GC_REDIRECT_TO_LOCAL
#include <gc/gc.h>
#include <string.h>
#include <stdio.h>
#include <sys/stat.h>

#include "int_funcs.h"
#include "types.h"

scm_value * list_reverse(scm_value * list)
{
	scm_value * result = SCM_CONST_NIL;
	while (SCM_TYPE(list) != NIL) {
		result = scm_value_new_pair(SCM_CAR(list), result);
		list = SCM_CDR(list);
		
	}
	return result;
}

char * load_file(char * filename) 
{
	struct stat stats;
	char * result;
	FILE * f = fopen(filename, "r");

	if (f == NULL) return NULL;	
	stat(filename, &stats);
	result = (char *) GC_MALLOC_ATOMIC(stats.st_size + 1);
	fread(result, 1, stats.st_size, f);
	fclose(f);
	
	return result; 
}

int is_list(scm_value * val)
{
	if (SCM_TYPE(val) == NIL) return 1;
	if (SCM_TYPE(val) == PAIR) return is_list(SCM_CDR(val));
	return 0;
}

int list_length(scm_value * list)
{
	int result = 0;
	while (SCM_TYPE(list) != NIL) {
		list = SCM_CDR(list);
		result++;
	}
	return result;
}

static void display_list(scm_value * list) 
{
	printf("(");
	while (SCM_TYPE(list) != NIL) {
		display_value(SCM_CAR(list));
		list = SCM_CDR(list);
		if (SCM_TYPE(list) != NIL) printf(" ");
	}
	printf(")");
}

static void display_value_generic(scm_value * val, int is_debug)
{
	int i;
	if (!is_debug && (is_list(val))) {
		display_list(val);
		return;
	}
	switch (SCM_TYPE(val)) {
		case LAMBDA:
			printf("#<Lambda "); 
			display_value(val->value.lambda->arg_list);
			printf(">");
			break;
		case SYMBOL:
			if (is_debug) printf("#<Sym (%s)>", SCM_SYMBOL(val));
			else printf("%s", SCM_SYMBOL(val));
			break;
		case INT:
			if (is_debug) printf("#<Int (%i)>", SCM_INT(val));
			else printf("%i", SCM_INT(val));
			break;
		case FLOAT:
			if (is_debug) printf("#<Float (%f)>", SCM_FLOAT(val));
			else printf("%f", SCM_FLOAT(val));
			break;

		case BOOL:
			if (is_debug) printf("#<Bool (%s)>", SCM_BOOL(val) ? "#t" : "#f");
			else printf("%s", SCM_BOOL(val) ? "#t" : "#f");
			break;
		case FUNC:
			printf("#<Func>");
			break;
		case SPEC_FORM:
			printf("#<Spec>");
			break;
		case VOID:
			printf("#<Void>");
			break;
		case MACRO:
			printf("#<Macro "); 
			display_value(val->value.lambda->arg_list);
			printf(">");
			break;
		case NIL:
			printf("()");
			break;
		case ERROR:
			printf("#<Err (%s)>", SCM_ERROR(val));
			break;
		case PAIR:
			printf("(");
			display_value_generic(SCM_CAR(val), is_debug);
			printf(" . ");
			display_value_generic(SCM_CDR(val), is_debug);
			printf(")");
			break;
		case CONTINUATION:
			printf("#<Continuation>");
			break;
		case VECTOR:
			if (is_debug) printf("#<Vector (");
			else printf("#(");
			for (i = 0; i < SCM_VECTOR(val)->size; i++) {
				display_value_generic(SCM_VECTOR(val)->items[i], is_debug);
				if (i < (SCM_VECTOR(val)->size - 1)) printf(" ");
			}
			printf(")");
			break;
	}
}

void display_value(scm_value * val)
{
	display_value_generic(val, 0);
}

void display_value_dbg(scm_value * val)
{
	display_value_generic(val, 1);
}
