%{
#include <gc/gc.h>
#include <string.h>
#include "scheme.tab.h"

%}

INITIAL_CHAR	[[:alnum:]]|\+|\*|\-|\\|=|\?|\!|<|>|_|\/|~|:|%
INNER_CHAR	[[:alnum:]]|\+|\*|\-|\\|=|\?|\!|<|>|_|\/|~|:|%|'|,|@|`|\.
LEFT_PAREN	"("
RIGHT_PAREN	")"
LEFT_BRACKET	"["
RIGHT_BRACKET	"]"
WHITE_SPACE 	[[:space:]]
QUOTE		"'"
UNQUOTE		","
UNQUOTE_SPLICING "@"
QUASIQUOTE	"`"
DOT		"."
SEMICOL		";"
NUMBER		[0-9]
MINUS		-
HASH		#


%%

{MINUS}{0,1}{NUMBER}+ 	{ 
	yylval.symbol = GC_MALLOC(strlen(yytext) + 1);
	strcpy(yylval.symbol, yytext);
	return T_NUMBER;
}

[+-]?[0-9]+(\.[0-9]+)?([eE][+-]?[0-9]+)? 	{ 
	yylval.symbol = GC_MALLOC(strlen(yytext) + 1);
	strcpy(yylval.symbol, yytext);
	return T_FLOAT_NUMBER;
}

{INITIAL_CHAR}{INNER_CHAR}* {
	yylval.symbol = GC_MALLOC(strlen(yytext) + 1);
	strcpy(yylval.symbol, yytext);
	return T_SYMBOL;
}

{LEFT_PAREN} 	{ return T_LEFT_PAREN; }
{RIGHT_PAREN} 	{ return T_RIGHT_PAREN; }
{LEFT_BRACKET} 	{ return T_LEFT_BRACKET; }
{RIGHT_BRACKET} { return T_RIGHT_BRACKET; }
{SEMICOL}.*$	/* ignores comments */
{WHITE_SPACE}+	/* ignores white space*/
{QUOTE}		{ return T_QUOTE; }
{UNQUOTE}	{ return T_UNQUOTE; }
{QUASIQUOTE}	{ return T_QUASIQUOTE; }
{UNQUOTE}{UNQUOTE_SPLICING}	{ return T_UNQUOTE_SPLICING; }
{DOT}		{ return T_DOT; }
{HASH}t		{ return T_TRUE; }
{HASH}f		{ return T_FALSE; }
{HASH}		{ return T_VECTOR; }
%%
