;;
;;
;; Basic functions
;;
;;


;;
;; List functions
;;
(define (map fun lst) 
  (if (null? lst) 
    lst 
    (cons (fun (car lst)) (map fun (cdr lst)))))

(define (apply func ls)
  (if (null? ls) (func) (func (car ls) (apply func (cdr ls)))))

(define (length x) 
  (if (null? x) 
    0 
    (+ 1 (length (cdr x)))))
			
(define (caar x) (car (car x)))
(define (cadr x) (car (cdr x)))
(define (cdar x) (cdr (car x)))
(define (cddr x) (cdr (cdr x)))
(define (caaar x) (car (car (car x))))
(define (cadar x) (car (cdr (car x))))
(define (caadr x) (car (car (cdr x))))
(define (cddar x) (cdr (cdr (car x))))
(define (caddr x) (car (cdr (cdr x))))
(define (cdadr x) (cdr (car (cdr x))))
(define (cdddr x) (cdr (cdr (cdr x))))

(define (list-tail ls k)
    (if (= k 0) ls (list-tail (cdr ls) (- k 1))))

(define (list-ref ls k)
    (car (list-tail ls k)))

(define (reverse ls)
    (define (iter source target)
          (if (null? source) target
	          (iter (cdr source) (cons (car source) target))))
      (iter ls ()))


(define (<= x y) (not (> x y)))
(define (>= x y) (not (< x y)))

;(define (append ls1 ls2) 
;	(define (app1 ls1 ls2) 
;		(if (null? ls1) 
;			ls2 
;			(app1 (cdr ls1) (cons (car ls1) ls2)))) 
;	(app1 (app1 ls1 '()) ls2))

(define (displayn arg)
	(display arg)
	(newline))

;;
;; Quotations
;;
;(define-macro (quasiquote ls) 
;	(define (unq la) 
;		(if (pair? la) 
;			(if (pair? (car la)) 
;				(if (equal? (caar la) 'unquote) 
;					(list 'cons (cadar la) (unq (cdr la))) 
;					(if (equal? (caar la) 'unquote-splicing) 
;						(list 'append (cadar la) (unq (cdr la))) 
;						(list 'cons (unq (car la)) (unq (cdr la))))) 
;				(list 'cons (unq (car la)) (unq (cdr la)))) 
;			(list 'quote la)))
;	(unq ls))

;;
;; Lambda extensions
;;
;(define-macro let (lambda args 
;	`((lambda ,(map car (car args)) ,@(cdr args)) ,@(map cadr (car args)))))

(define-macro (let* assgn . body)
	(if (null? assgn)
		`((lambda () ,@body))
			`(let (,(car assgn))
				(let* ,(cdr assgn)
					,@body))))

;(define-macro (letrec assgn . body)
(define-macro (letrec assgn . body)
	      `((lambda ()
		  ,@(map (lambda (i)
			   `(define ,(car i) ,(cadr i)))
			 assgn)
		  ,@body)))
	
;;
;; Conditions
;;

(define-macro cond
	      (lambda clist
		(let ((symbol (gensym)))
		  (if (null? clist)
		    `(if #f #f) ; FIXME should return Void in better way
		    (if (equal? (caar clist) 'else)
		      `(begin ,@(cdar clist))
		      `(let ((,symbol ,(caar clist)))
			 (if ,symbol
			   ,(if (null? (cdar clist))
			      symbol
			      (if (equal? (cadar clist) '=>)
				`(,(caddar clist) ,symbol)
				`(begin ,@(cdar clist))))
			   (cond ,@(cdr clist)))))))))


;(define-macro and (lambda args
;	(if (null? args)
;		#t
;		(if (null? (cdr args))
;			(car args)
;			`(if ,(car args)
;				(and ,@(cdr args))
;				#f)))))

;(define-macro or (lambda args
;	(if (null? args)
;		#f
;		(if (null? (cdr args))
;			(car args)
;			(let ((result (gensym))) 
;				`(let ((,result ,(car args)))
;					(if ,result
;					,result
;					(or ,@(cdr args)))))))))
