;;
;; Implementace hopfieldovy site
;;
;; (c) 2006 Petr Krajca, krajcap@inf.upol.cz
;;

;;
;; Pomocne funkce pro praci s matici
;;

;; vytvori ctvercovou matici dane dimenze
(define (matrix-create dimension)
	(define (build-list items build-func)
		(if (= items 0) '()
			(cons (build-func) (build-list (- items 1) build-func))))
	(build-list dimension
		(lambda () 
			(build-list dimension (lambda () 0)))))

;; vraci radek z matice (pomocna fce.)
(define (matrix-get-row matrix row)
	(list-ref matrix row))

;; pricte hodnotu k prvku specifikovanemu radkem a sloupcem hodnotu val
(define (matrix-add! matrix row col val)
	(let 
		((item (list-tail (matrix-get-row matrix row) col)))
		(set-car! item (+ (car item) val))))

;; vraci hodnotu z matice
(define (matrix-get matrix row col)
	(list-ref (matrix-get-row matrix row) col))

;; vypise matici
(define (matrix-print! matrix)
	(map (lambda (x) (begin (display x) (newline))) matrix)
	(newline))

;;
;;
;; Implementace algoritmu
;;
;;

;; Pomocna fce. obdoba standardni map s tim rozdilem, ze jako func.
;; akceptuje funkci dvou promennych. kde se v prvnim parametru predava
;; prvek ze seznamu a ve druhem poradi prvku v seznamu (indexovano od 0)
(define (counting-map func l)
	(define (walk l position)
		(if (null? l) l
			(cons 
				(func (car l) position) 
				(walk (cdr l) (+ position 1)))))
	(walk l 0))

;; Nauci sit reprezentovanou matici vzor dany seznamem
(define (hopfield-train! matrix pattern)
	(counting-map (lambda (p1 pos1)
		(counting-map (lambda (p2 pos2)
			(if (not (= pos1 pos2))	
				(matrix-add! matrix pos1 pos2 (if (= p1 p2) 1 -1))))
		pattern)) pattern))

;; Upravi vzor tak, aby se priblizil nekteremu z naucenych
(define (hopfield-run matrix pattern)
	(counting-map (lambda (p1 pos1)
		(let 
			((neuron-state (apply + (counting-map 
				(lambda (p2 pos2)
					(* p2 (matrix-get matrix pos1 pos2))) pattern))))
			(if (>= neuron-state 0) 1 -1))) 
		pattern))
		
;; Provadi vybavovaci smycku dokud se vzor meni
;; lze dodat funkci se tremi parametry, pro sledovani prubehu vypoctu
(define (hopfield-cycle matrix pattern . introspect)
	(define (iter previous current cycle)
		(if (not (null? introspect)) ((car introspect) matrix current cycle))
		(if (equal? previous current) current
			(iter
				current
				(hopfield-run matrix current)
				(+ cycle 1))))
	(iter '() pattern 0))

;; vraci energii site
(define (hopfield-energy matrix pattern)
	(* -1 (apply + 
		(counting-map (lambda (p1 pos1)
			(apply + (counting-map 
				(lambda (p2 pos2) (* p1 p2 (matrix-get matrix pos1 pos2)))
				  pattern)))
			pattern))))

;; funkce na introspekci vybavovani
(define (introspect matrix pattern cycle)
	(display "Cycle: ")
	(display cycle)
	(newline)
	(display "Energy: ")
	(display (hopfield-energy matrix pattern))
	(newline)
	(matrix-print! (list->matrix 5 (map c-1>0 pattern)))
	(newline)
	(newline))
