#ifndef STACK_C
#define STACK_C

#define GC_REDIRECT_TO_LOCAL
#include <gc/gc.h>

#include "stack.h"

#define INITIAL_CAPACITY	16
static inline struct stack * stack_new()
{
	struct stack * result = (struct stack *) GC_MALLOC(sizeof(struct stack));
	result->capacity = INITIAL_CAPACITY;
	result->size = 0;
	result->data = (void **) GC_MALLOC(sizeof(void *) * result->capacity);
	return result;
}

static inline void stack_push(struct stack * stack, void * value)
{
	if (stack->size == stack->capacity) {
		stack->capacity *= 2;
		stack->data = (void **) GC_realloc(stack->data, sizeof(void *) * stack->capacity);
	}
	stack->data[stack->size++] = value;
}

static inline void * stack_pop(struct stack * stack)
{
	if (stack->size == 0) return NULL;
	void * result = stack->data[stack->size - 1];
	stack->data[--stack->size] = NULL;
	return result;
}

static inline void * stack_top(struct stack * stack)
{
	if (stack->size == 0) return NULL;
	return stack->data[stack->size - 1];
}


static inline int stack_empty(struct stack * stack)
{
	return (stack->size == 0);
}

static inline stack_item stack_internal_get_top(struct stack * stack)
{
	return stack->size;
}

static inline stack_item stack_internal_get_bottom(struct stack * stack)
{
	return 1;
}

static inline stack_item stack_internal_get_next(struct stack * stack, stack_item item)
{
	return item - 1;
}

static inline stack_item stack_internal_get_prev(struct stack * stack, stack_item item)
{
	return item + 1;
}

static inline stack_item stack_internal_get_done(struct stack * stack, stack_item item)
{
	return ((item <= 0) || (item > stack->size));
}

static inline void * stack_internal_get_value(struct stack * stack, stack_item item)
{
	return stack->data[item - 1];
}


static inline void stack_append(struct stack * s1, struct stack * s2)
{
	
//	mutex_t lck;
//	mutex_init(lck);
//	mutex_lock(lck); /* memcpy - threadsafe */
	s1->data = GC_realloc(s1->data, sizeof(void *) * (s1->size + s2->size));
	s1->capacity = (s1->size + s2->size + 1);

	memcpy(s1->data + s1->size, s2->data, sizeof(void *) * s2->size);
	s2->capacity = INITIAL_CAPACITY;
	s2->data = (void **) GC_MALLOC(sizeof(void *) * s2->capacity);

	s1->size = s1->size + s2->size;
	s2->size = 0;
//	mutex_unlock(lck);
}

#endif
