#ifndef TYPES_H
#define TYPES_H

#define SCM_CONST_NIL (scm_const_nil)
#define SCM_CONST_VOID	(scm_const_void)
#define SCM_CONST_FALSE	(scm_const_false)
#define SCM_CONST_TRUE	(scm_const_true)

#define INTEGER_POINTER_HACK 
#ifndef INTEGER_POINTER_HACK
	#define SCM_TYPE(x)	(x->type)
	#define SCM_INT(x)	(x->value.integer)
#else
	#define SCM_TYPE(x)	((((long)(x)) & 1) ? INT : x->type)
	#define	SCM_INT(x)	(int)((long)x >> 1)
#endif

#define SCM_BOOL(x)	(x->value.bool)
#define SCM_SYMBOL(x)	(x->value.symbol)
#define SCM_FLOAT(x)	(x->value.flt)
#define SCM_PAIR(x)	(x->value.pair)
#define SCM_VECTOR(x)	(x->value.vector)
#define SCM_CAR(x)	(x->value.pair.ar)
#define SCM_CDR(x)	(x->value.pair.dr)
#define SCM_FUNC(x)	(x->value.func)
#define SCM_SPEC_FORM(x)	(x->value.spec_form)
#define SCM_LAMBDA(x)	(x->value.lambda)
#define SCM_ERROR(x)	(x->value.err_desc)
#define SCM_IS_NULL(x)	(SCM_TYPE(x) == NIL)
#define SCM_CONS(x,y)	(scm_value_new_pair(x, y))

#define SCM_CAAR(x)	(SCM_CAR(SCM_CAR(x)))
#define SCM_CADR(x)	(SCM_CAR(SCM_CDR(x)))
#define SCM_CDDR(x)	(SCM_CDR(SCM_CDR(x)))
#define SCM_CDAR(x)	(SCM_CDR(SCM_CAR(x)))
#define SCM_CADDR(x)	(SCM_CAR(SCM_CDR(SCM_CDR(x))))
#define SCM_CDDDR(x)	(SCM_CDR(SCM_CDR(SCM_CDR(x))))
#define SCM_CDDAR(x)	(SCM_CDR(SCM_CDR(SCM_CAR(x))))
#define SCM_CADAR(x)	(SCM_CAR(SCM_CDR(SCM_CAR(x))))
#define SCM_CDDDDR(x)	(SCM_CDR(SCM_CDR(SCM_CDR(SCM_CDR(x)))))
#define SCM_CADDAR(x)	(SCM_CAR(SCM_CDR(SCM_CDR(SCM_CAR(x)))))
#define SCM_CAADAR(x)	(SCM_CAR(SCM_CAR(SCM_CDR(SCM_CAR(x)))))

struct scm_value;
struct lambda;
struct scm_vm;

#include "env.h"

typedef struct scm_value * (*scm_func)(int cnt, struct scm_value **);
typedef struct scm_value * (*scm_spec_form)(int cnt, struct scm_value **, scm_env *);

typedef enum scm_type {
	INT,
	FLOAT,
	PAIR,
	SYMBOL,
	FUNC,
	SPEC_FORM,
	LAMBDA,
	MACRO,
	BOOL,
	NIL,
	VOID,
	ERROR,
	CONTINUATION,
	PRECOMPILED_MACRO,
	VECTOR
} scm_type;

typedef enum scm_spec_type {
	SP_DEFINE, 
	SP_LAMBDA,
	SP_SET,
	SP_QUOTE,
	SP_MACRO,
	SP_IF,
	SP_CALLCC,
	SP_SETCAR,
	SP_SETCDR,
	SP_VECTOR_SET,
	SP_MAP_EXP,
	SP_DISPLAY,
	SP_NEWLINE
} scm_spec_type;

typedef struct scm_vector {
	int size;
	struct scm_value ** items;
} scm_vector;

typedef struct scm_value {
	scm_type type;
	union {
		int integer;
		double flt;
		int bool;
		char * symbol;
		struct scm_pair {
			struct scm_value * ar;
			struct scm_value * dr;
		} pair;
		scm_vector * vector;
		scm_func func;
		scm_spec_type spec_form;
		struct lambda * lambda;
		char * err_desc;
		struct scm_vm * continuation;
	} value;
} scm_value;

struct lambda {
	scm_env * parent;
	scm_value * arg_list;
	scm_value * define_list;
	scm_value * cmd_list;
	unsigned int local_symbols;
};

extern scm_value * scm_const_nil;
extern scm_value * scm_const_true;
extern scm_value * scm_const_false;
extern scm_value * scm_const_void;

void scm_init_constants();
#ifndef INTEGER_POINTER_HACK
scm_value * scm_value_new_int(int val);
#else
#define scm_value_new_int(__val) ((scm_value *)(long)(1 | ((__val) << 1))) 
#endif
scm_value * scm_value_new_symbol(char * symbol);
scm_value * scm_value_new_float(double val);
scm_value * scm_value_new_pair(scm_value * ar, scm_value * dr);
scm_value * scm_value_new_nil();
scm_value * scm_value_new_func(scm_func func);
scm_value * scm_value_new_spec_form(scm_spec_type type);
scm_value * scm_value_new_precompiled_macro(scm_func type);
scm_value * scm_value_new_lambda(scm_env * env, scm_value * arg_list, scm_value * cmd_list);
scm_value * scm_value_new_lambda_clone(scm_value * val);
scm_value * scm_value_new_bool(int value); 
scm_value * scm_value_new_void(); 
scm_value * scm_value_new_error(char * desc); 
scm_value * scm_value_new_continuation(struct scm_vm * continuation);
scm_value * scm_value_new_vector(int count, scm_value * val);
scm_value * scm_value_new_vector_from_list(scm_value * items);

#endif
