#ifndef VM_H
#define VM_H

#include <gc/gc.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>

#include "types.h"
#include "mutex.h"
#include "env.h"
#include "rt.h"
#include "stack.h"

#define NOTAIL	(0)
#define	TAIL	(1)
#define HEAD	(2)

enum command_op {
	OP_EVAL, 
	OP_INSPECT, 
	OP_RUN, 
	OP_DEFINE, 
	OP_DEFINE_MACRO,
	OP_IF,
	OP_SET,
	OP_DISCARD,
	OP_RETURN,
	OP_FEVAL,
	OP_WAIT,
	OP_DROP,
	OP_APPLY_CONTINUATION,
	OP_BLOCK,
	OP_SETCAR,
	OP_SETCDR,
	OP_VECTOR_SET,
	OP_MAP_EXP,
	OP_DISPLAY,
	OP_NEWLINE
};

struct scm_vm;

typedef struct command {
	enum command_op op;	
	scm_value * arg1; 
	scm_env *env;
	struct scm_vm * vm;
	struct command * next_pooled;
	unsigned char tail_position;
} command;

typedef struct scm_vm {
	scm_rt * runtime;
	struct stack * exct;
	struct stack * rslt;	
	struct scm_vm * parent;
	mutex_t lock;
	command * command_pool;
	unsigned long * vm_id;
} scm_vm;

static inline scm_vm * vm_new(scm_rt * runtime);
static inline scm_vm * vm_new_init(scm_rt * runtime, command * initial_command);
static inline int vm_done(scm_vm * vm);
static inline void vm_push_command(scm_vm * vm, enum command_op op, scm_env * env, scm_value * arg);
static inline void vm_push_command_tail(scm_vm * vm, enum command_op op, scm_env * env, scm_value * arg, unsigned char tail_position);
static inline command * vm_pop_command(scm_vm * vm);
static inline scm_value * vm_result_top(scm_vm * vm);
static inline scm_value * vm_result_pop(scm_vm * vm);
static inline void vm_result_push(scm_vm * vm, scm_value * value);
static inline void vm_join(scm_vm * vm, scm_vm * slave);
static inline void vm_lock(scm_vm * vm);
static inline void vm_unlock(scm_vm * vm);
static inline void vm_drop_fevals(scm_vm * vm);
static inline scm_rt * vm_get_runtime(scm_vm * vm);
static inline void vm_command_free(scm_vm * vm, command * cmd);
static inline command * vm_command_new(scm_vm * vm, enum command_op op, scm_env * env, scm_value * arg, unsigned char tail_position);
static inline void vm_error(scm_vm * vm, scm_value * error);

#endif
